# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import json
import os
import pathlib
import platform
import shutil
from pathlib import Path
from unittest import mock

import pytest
from callee import Contains, Matching

from mozphab import exceptions, mozphab

from .conftest import git_out, search_diff, search_rev

call_conduit = mock.Mock()


def by_line_mock(*args, **_kwargs):
    # join args to catch unicode errors
    " ".join(*args)
    return ["Revision URI: http://example.test/D123"]


check_call_by_line = mock.Mock()
check_call_by_line.side_effect = by_line_mock

initial_sha = None


def test_submit_create(in_process, git_repo_path: pathlib.Path, init_sha):
    call_conduit.reset_mock()
    call_conduit.side_effect = (
        # ping
        {},
        # diffusion.repository.search
        {"data": [{"phid": "PHID-REPO-1", "fields": {"vcs": "git"}}]},
        # user search
        [{"userName": "alice", "phid": "PHID-USER-1"}],
        # differential.creatediff
        {"phid": "PHID-DIFF-1", "diffid": "1"},
        # differential.revision.edit
        {"object": {"id": "123", "phid": "PHID-DREV-123"}},
        # differential.setdiffproperty
        {},
    )
    (git_repo_path / "X").write_text("ą\nb\nc\n", encoding="utf-8")
    (git_repo_path / "Y").write_text("no line ending")
    git_out("add", ".")
    (git_repo_path / "msg").write_text("Ą r?alice", encoding="utf-8")
    git_out("commit", "--file", "msg")
    (git_repo_path / "untracked").write_text("a\n")

    mozphab.main(["submit", "--yes", "--bug", "1", init_sha], is_development=True)

    log = git_out("log", "--format=%s%n%n%b", "-1")
    expected = """
Bug 1 - Ą r?alice

Differential Revision: http://example.test/D123

"""
    assert log.strip() == expected.strip()
    assert mock.call("conduit.ping", {}) in call_conduit.call_args_list
    assert (
        mock.call("user.query", {"usernames": ["alice"]}) in call_conduit.call_args_list
    )
    assert (
        mock.call(
            "diffusion.repository.search",
            {"limit": 1, "constraints": {"callsigns": ["TEST"]}},
        )
        in call_conduit.call_args_list
    )
    assert (
        mock.call(
            "differential.creatediff",
            {
                "sourceControlPath": "/",
                "sourceControlSystem": "git",
                "lintStatus": "none",
                "sourcePath": mock.ANY,
                "unitStatus": "none",
                "sourceMachine": "http://example.test",
                "sourceControlBaseRevision": mock.ANY,
                "repositoryPHID": "PHID-REPO-1",
                "branch": "HEAD",
                "changes": [
                    {
                        "commitHash": mock.ANY,
                        "awayPaths": [],
                        "newProperties": {"unix:filemode": "100644"},
                        "oldPath": None,
                        "hunks": [
                            {
                                "oldOffset": 0,
                                "oldLength": 0,
                                "newOffset": 1,
                                "newLength": 3,
                                "addLines": 3,
                                "delLines": 0,
                                "corpus": "+ą\n+b\n+c\n",
                                "isMissingOldNewline": False,
                                "isMissingNewNewline": False,
                            }
                        ],
                        "oldProperties": {},
                        "currentPath": "X",
                        "fileType": 1,
                        "type": 1,
                        "metadata": {},
                    },
                    {
                        "commitHash": mock.ANY,
                        "awayPaths": [],
                        "newProperties": {"unix:filemode": "100644"},
                        "oldPath": None,
                        "hunks": [
                            {
                                "oldOffset": 0,
                                "oldLength": 0,
                                "newOffset": 1,
                                "newLength": 1,
                                "addLines": 1,
                                "delLines": 0,
                                "corpus": (
                                    "+no line ending\n\\ No newline at end of file\n"
                                ),
                                "isMissingOldNewline": False,
                                "isMissingNewNewline": True,
                            }
                        ],
                        "oldProperties": {},
                        "currentPath": "Y",
                        "fileType": 1,
                        "type": 1,
                        "metadata": {},
                    },
                ],
                "creationMethod": "moz-phab-git",
            },
        )
        in call_conduit.call_args_list
    )
    assert (
        mock.call(
            "differential.setdiffproperty",
            {"diff_id": "1", "name": "local:commits", "data": ~Contains('"rev":')},
        )
        in call_conduit.call_args_list
    )
    call_conduit.assert_any_call(
        "differential.revision.edit",
        {
            "transactions": [
                {"type": "title", "value": "Bug 1 - Ą r?alice"},
                {"type": "summary", "value": ""},
                {"type": "reviewers.set", "value": ["PHID-USER-1"]},
                {"type": "bugzilla.bug-id", "value": "1"},
                {"type": "update", "value": "PHID-DIFF-1"},
            ]
        },
    )


def test_submit_create_added_not_commited(
    in_process, git_repo_path: pathlib.Path, init_sha
):
    call_conduit.side_effect = (
        # ping
        {},
        # diffusion.repository.search
        {"data": [{"phid": "PHID-REPO-1", "fields": {"vcs": "git"}}]},
        # user search
        [{"userName": "alice", "phid": "PHID-USER-1"}],
        # differential.creatediff
        {"phid": "PHID-DIFF-1", "diffid": "1"},
        # differential.revision.edit
        {"object": {"id": "123", "phid": "PHID-DREV-123"}},
        # differential.setdiffproperty
        {},
    )
    (git_repo_path / "X").write_text("ą\r\nb\nc\n", encoding="utf-8")
    (git_repo_path / "Y").write_text("no line ending")
    git_out("add", ".")
    (git_repo_path / "msg").write_text("Ą r?alice", encoding="utf-8")
    git_out("commit", "--file", "msg")
    (git_repo_path / "untracked").write_text("a\n")
    git_out("add", "untracked")

    with pytest.raises(exceptions.Error) as excinfo:
        mozphab.main(["submit", "--yes", "--bug", "1", init_sha], is_development=True)

    assert "Uncommitted changes present." in str(excinfo.value)


def test_submit_create_no_bug(in_process, git_repo_path: pathlib.Path, init_sha):
    call_conduit.reset_mock()
    call_conduit.side_effect = (
        # ping
        {},
        # diffusion.repository.search
        {"data": [{"phid": "PHID-REPO-1", "fields": {"vcs": "git"}}]},
        # user search
        [{"userName": "alice", "phid": "PHID-USER-1"}],
        # differential.creatediff
        {"phid": "PHID-DIFF-1", "diffid": "1"},
        # differential.revision.edit
        {"object": {"id": "123", "phid": "PHID-DREV-123"}},
        # differential.setdiffproperty
        {},
    )
    testfile = git_repo_path / "X"
    testfile.write_text("a\n")
    git_out("add", ".")
    msgfile = git_repo_path / "msg"
    msgfile.write_text("A r?alice")
    git_out("commit", "--file", "msg")

    mozphab.main(["submit", "--yes", "--no-bug", init_sha], is_development=True)

    log = git_out("log", "--format=%s%n%n%b", "-1")
    expected = """
A r?alice

Differential Revision: http://example.test/D123
"""
    assert log.strip() == expected.strip()


def test_submit_create_binary(
    in_process, git_repo_path: pathlib.Path, init_sha, data_file
):
    call_conduit.reset_mock()
    call_conduit.side_effect = (
        # ping
        {},
        # diffusion.repository.search
        {"data": [{"phid": "PHID-REPO-1", "fields": {"vcs": "git"}}]},
        # file.allocate
        {"filePHID": None, "upload": True},
        # file.upload
        {},
        # differential.creatediff
        {"phid": "PHID-DIFF-1", "diffid": "1"},
        # differential.revision.edit
        {"object": {"id": "123", "phid": "PHID-DREV-123"}},
        # differential.setdiffproperty
        {},
    )
    shutil.copyfile(str(data_file), str(git_repo_path / "img.png"))
    git_out("add", ".")
    git_out("commit", "-m", "IMG")

    mozphab.main(["submit", "--yes", "--bug", "1", init_sha], is_development=True)

    log = git_out("log", "--format=%s%n%n%b", "-1")
    expected = """
Bug 1 - IMG

Differential Revision: http://example.test/D123
"""
    assert log.strip() == expected.strip()
    assert (
        mock.call(
            "file.allocate",
            {"name": "img.png", "contentHash": mock.ANY, "contentLength": 182},
        )
        in call_conduit.call_args_list
    )
    assert (
        mock.call("file.upload", {"data_base64": mock.ANY, "name": "img.png"})
        in call_conduit.call_args_list
    )


def test_submit_create_binary_existing(
    in_process, git_repo_path: pathlib.Path, init_sha, data_file
):
    call_conduit.reset_mock()
    call_conduit.side_effect = (
        # ping
        {},
        # diffusion.repository.search
        {"data": [{"phid": "PHID-REPO-1", "fields": {"vcs": "git"}}]},
        # file.allocate
        {"filePHID": "PHID-FILE-1", "upload": False},
        # no file.upload call
        # differential.creatediff
        {"phid": "PHID-DIFF-1", "diffid": "1"},
        # differential.revision.edit
        {"object": {"id": "123", "phid": "PHID-DREV-123"}},
        # differential.setdiffproperty
        {},
    )
    shutil.copyfile(str(data_file), str(git_repo_path / "img.png"))
    git_out("add", ".")
    git_out("commit", "-m", "IMG")

    mozphab.main(["submit", "--yes", "--bug", "1", init_sha], is_development=True)

    log = git_out("log", "--format=%s%n%n%b", "-1")
    expected = """
Bug 1 - IMG

Differential Revision: http://example.test/D123
"""
    assert log.strip() == expected.strip()
    assert (
        mock.call(
            "file.allocate",
            {"name": "img.png", "contentHash": mock.ANY, "contentLength": 182},
        )
        in call_conduit.call_args_list
    )
    assert mock.call("file.upload", mock.ANY) not in call_conduit.call_args_list


def test_submit_create_binary_chunked(
    in_process, git_repo_path: pathlib.Path, init_sha, data_file
):
    call_conduit.reset_mock()
    call_conduit.side_effect = (
        # ping
        {},
        # diffusion.repository.search
        {"data": [{"phid": "PHID-REPO-1", "fields": {"vcs": "git"}}]},
        # file.allocate
        {"filePHID": "PHID-FILE-1", "upload": True},
        # file.querychunks
        [
            {"byteStart": "0", "byteEnd": "4194304", "complete": False},
            {"byteStart": "4194304", "byteEnd": "8388608", "complete": False},
            {"byteStart": "8388608", "byteEnd": "8425160", "complete": False},
        ],
        # file.uploadchunk
        {},
        # file.uploadchunk
        {},
        # file.uploadchunk
        {},
        # differential.creatediff
        {"phid": "PHID-DIFF-1", "diffid": "1"},
        # differential.revision.edit
        {"object": {"id": "123", "phid": "PHID-DREV-123"}},
        # differential.setdiffproperty
        {},
    )
    shutil.copyfile(str(data_file), str(git_repo_path / "img.png"))
    git_out("add", ".")
    git_out("commit", "-m", "IMG")

    mozphab.main(["submit", "--yes", "--bug", "1", init_sha], is_development=True)

    log = git_out("log", "--format=%s%n%n%b", "-1")
    expected = """
Bug 1 - IMG

Differential Revision: http://example.test/D123
"""
    assert log.strip() == expected.strip()
    assert (
        mock.call(
            "file.allocate",
            {"name": "img.png", "contentHash": mock.ANY, "contentLength": 182},
        )
        in call_conduit.call_args_list
    )
    assert (
        mock.call("file.querychunks", {"filePHID": "PHID-FILE-1"})
        in call_conduit.call_args_list
    )
    assert (
        mock.call(
            "file.uploadchunk",
            {
                "filePHID": "PHID-FILE-1",
                "byteStart": 0,
                "data": mock.ANY,
                "dataEncoding": "base64",
            },
        )
        in call_conduit.call_args_list
    )
    assert (
        mock.call(
            "file.uploadchunk",
            {
                "filePHID": "PHID-FILE-1",
                "byteStart": 4194304,
                "data": mock.ANY,
                "dataEncoding": "base64",
            },
        )
        in call_conduit.call_args_list
    )
    assert (
        mock.call(
            "file.uploadchunk",
            {
                "filePHID": "PHID-FILE-1",
                "byteStart": 8388608,
                "data": mock.ANY,
                "dataEncoding": "base64",
            },
        )
        in call_conduit.call_args_list
    )


def test_submit_create_no_checkout(in_process, git_repo_path: pathlib.Path, init_sha):
    """Tests that updates which rewrite the commit message:

    1. do
    2. update all branches containing the rewritten commits
    3. don't update the mtime of touched files
    """
    call_conduit.reset_mock()
    call_conduit.side_effect = (
        # ping
        {},
        # diffusion.repository.search
        {"data": [{"phid": "PHID-REPO-1", "fields": {"vcs": "git"}}]},
        # user search
        [{"userName": "alice", "phid": "PHID-USER-1"}],
        # First diff
        # differential.creatediff
        {"phid": "PHID-DIFF-1", "diffid": "1"},
        # differential.revision.edit
        {"object": {"id": "123", "phid": "PHID-DREV-123"}},
        # differential.setdiffproperty
        {},
        # Second diff
        # differential.creatediff
        {"phid": "PHID-DIFF-2", "diffid": "2"},
        # differential.revision.edit
        {"object": {"id": "124", "phid": "PHID-DREV-124"}},
        # differential.setdiffproperty
        {},
    )

    git_out("checkout", "-qb", "first")

    (git_repo_path / "X").write_text("ą\nb\nc\n", encoding="utf-8")
    (git_repo_path / "Y").write_text("no line ending")
    git_out("add", ".")
    (git_repo_path / "msg").write_text("Ą r?alice", encoding="utf-8")
    git_out("commit", "--file", "msg")
    (git_repo_path / "msg").unlink()

    (git_repo_path / "Z").write_text("ą\nb\nc\n", encoding="utf-8")
    git_out("add", ".")
    (git_repo_path / "msg").write_text("B r?alice", encoding="utf-8")
    git_out("commit", "--file", "msg")
    (git_repo_path / "msg").unlink()

    first_branch_tip = git_out("rev-parse", "first")
    first_branch_tree = git_out("rev-parse", "first^{tree}")

    # Create a second branch sharing the bottom commit.
    git_out("checkout", "-qb", "second", "HEAD^")

    AA = git_repo_path / "AA"
    AA.write_text("ą\nb\nc\n", encoding="utf-8")
    git_out("add", ".")
    (git_repo_path / "msg").write_text("C r?alice", encoding="utf-8")
    git_out("commit", "--file", "msg")
    (git_repo_path / "msg").unlink()

    AA_old_mtime = AA.stat().st_mtime
    second_branch_tip = git_out("rev-parse", "second")
    second_branch_tree = git_out("rev-parse", "second^{tree}")

    start_branch = git_out("rev-parse", "--abbrev-ref", "HEAD").strip()

    mozphab.main(["submit", "--yes", "--bug", "1", init_sha], is_development=True)

    current_branch = git_out("rev-parse", "--abbrev-ref", "HEAD").strip()

    new_first_branch_tip = git_out("rev-parse", "first")
    new_first_branch_tree = git_out("rev-parse", "first^{tree}")
    new_second_branch_tip = git_out("rev-parse", "second")
    new_second_branch_tree = git_out("rev-parse", "second^{tree}")

    # Branches should have been rewritten, but trees should be unchanged.
    assert (
        new_first_branch_tip != first_branch_tip
    ), "The tip of the first branch should have changed."
    assert (
        new_first_branch_tree == first_branch_tree
    ), "The tree of the first branch should not have changed."
    # Ensure commits have been updated in other branches, too.
    first_log = git_out("log", "--format=%s%n%n%b", "-2", "first").strip()
    assert (
        "Differential Revision: http://example.test/D123" in first_log
    ), "Some commits in the first branch should contain the Differential Revision."
    assert (
        "D124" not in first_log
    ), "Unexpected Differential Revision found in first branch."

    assert (
        new_second_branch_tip != second_branch_tip
    ), "The tip of the second branch should have changed."
    assert (
        new_second_branch_tree == second_branch_tree
    ), "The tree of the second branch should not have changed."

    sep = "---sep---"
    second_log = (
        git_out("log", "--format=%s%n%n%b" + sep, "-2", "second").strip().split(sep)
    )
    assert (
        "Differential Revision: http://example.test/D123" in second_log[1]
    ), "The commits in the second branch should contain the first Differential Revision."
    assert (
        "Differential Revision: http://example.test/D124" in second_log[0]
    ), "The commits in the second branch should contain the second Differential Revision."

    # Files mtimes should not have changed.
    assert (
        AA.stat().st_mtime == AA_old_mtime
    ), "The mtime of file AA should not have changed."

    # Make sure we're still on the same branch
    assert current_branch == start_branch, "The current branch has changed"


def test_submit_update(in_process, git_repo_path: pathlib.Path, init_sha):
    call_conduit.reset_mock()
    call_conduit.side_effect = (
        # ping
        {},
        # diffusion.repository.search
        {"data": [{"phid": "PHID-REPO-1", "fields": {"vcs": "git"}}]},
        # diffusion.revision.search
        {"data": [search_rev(rev=123, reviewers=["alice"])]},
        # diffusion.diff.search
        {"data": [search_diff()]},
        # whoami
        {"phid": "PHID-USER-1"},
        # differential.creatediff
        {"phid": "PHID-DIFF-2", "diffid": "2"},
        # differential.revision.edit
        {"object": {"id": "123", "phid": "PHID-DREV-123"}},
        # differential.setdiffproperty
        {},
    )
    testfile = git_repo_path / "X"
    testfile.write_text("ą", encoding="utf-8")
    git_out("add", "X")
    msgfile = git_repo_path / "msg"
    msgfile.write_text(
        "Bug 1 - Ą r?alice\n\nDifferential Revision: http://example.test/D123",
        encoding="utf-8",
    )
    git_out("commit", "--file", "msg")

    mozphab.main(
        ["submit", "--yes", "--message", "update message ćwikła", init_sha],
        is_development=True,
    )

    log = git_out("log", "--format=%s%n%n%b", "-1").strip()
    expected = "Bug 1 - Ą r?alice\n\nDifferential Revision: http://example.test/D123"
    assert log == expected
    call_conduit.assert_any_call(
        "differential.creatediff",
        Matching(lambda x: x.get("repositoryPHID") == "PHID-REPO-1"),
    )
    call_conduit.assert_any_call(
        "differential.revision.edit",
        {
            "objectIdentifier": 123,
            "transactions": [
                {"type": "title", "value": "Bug 1 - Ą r?alice"},
                {"type": "summary", "value": ""},
                {"type": "comment", "value": "update message ćwikła"},
                {"type": "update", "value": "PHID-DIFF-2"},
            ],
        },
    )


def test_submit_update_uplift(in_process, git_repo_path: pathlib.Path, init_sha):
    call_conduit.reset_mock()
    call_conduit.side_effect = [
        # conduit.ping
        {},
        # diffusion.repository.search
        {"data": [{"phid": "PHID-REPO-1", "fields": {"vcs": "git"}}]},
        # differential.revision.search
        {"data": [search_rev(rev=1000, repo="PHID-REPO-BETA", reviewers=["#group"])]},
        # differential.diff.search
        {"data": [search_diff()]},
        # user.whoami
        {"phid": "PHID-USER-1"},
        # differential.creatediff
        {"phid": "PHID-DIFF-2", "diffid": "2"},
        # differential.revision.edit
        {"object": {"id": "1000", "phid": "PHID-DREV-1"}},
        # differential.setdiffproperty
        {},
    ]
    testfile = git_repo_path / "X"
    testfile.write_text("ą", encoding="utf-8")
    git_out("add", "X")
    msgfile = git_repo_path / "msg"
    msgfile.write_text(
        "Bug 1 - Ą\n\n"
        "Original Revision: http://example.test/D999\n\n"
        "Differential Revision: http://example.test/D1000",
        encoding="utf-8",
    )
    git_out("commit", "--file", "msg")

    mozphab.main(["submit", "-y", init_sha], is_development=True)

    log = git_out("log", "--format=%s%n%n%b", "-1").strip()
    expected = (
        "Bug 1 - Ą\n\n"
        "Original Revision: http://example.test/D999\n\n"
        "Differential Revision: http://example.test/D1000"
    )
    assert log == expected
    call_conduit.assert_any_call(
        "differential.creatediff",
        Matching(lambda x: x.get("repositoryPHID") == "PHID-REPO-BETA"),
    )
    call_conduit.assert_any_call(
        "differential.revision.edit",
        {
            "objectIdentifier": 1000,
            "transactions": [
                {"type": "title", "value": "Bug 1 - Ą"},
                {
                    "type": "summary",
                    "value": "\nOriginal Revision: http://example.test/D999",
                },
                {"type": "update", "value": "PHID-DIFF-2"},
            ],
        },
    )


def test_submit_update_no_change(
    in_process,
    git_repo_path: pathlib.Path,
    init_sha,
    git_sha,
    caplog: pytest.LogCaptureFixture,
):
    testfile = git_repo_path / "X"
    testfile.write_text("a")
    git_out("add", ".")
    msgfile = git_repo_path / "msg"
    msgfile.write_text(
        """\
Bug 1 - A, r=test

Differential Revision: http://example.test/D123
"""
    )
    git_out("commit", "--file", "msg")
    sha = git_sha()
    call_conduit.reset_mock()
    call_conduit.side_effect = (
        # ping
        {},
        # diffusion.repository.search
        {"data": [{"phid": "PHID-REPO-1", "fields": {"vcs": "git"}}]},
        # diffusion.revision.search
        {"data": [search_rev(rev=123, reviewers=("test",))]},
        # diffusion.diff.search
        {"data": [search_diff(node=sha)]},
        # whoami
        {"phid": "PHID-USER-1"},
    )

    mozphab.main(
        ["submit", "--yes"] + [init_sha],
        is_development=True,
    )

    assert "No changes to submit." in caplog.messages


def test_submit_remove_cr(in_process, git_repo_path: pathlib.Path, init_sha):
    call_conduit.reset_mock()
    call_conduit.side_effect = [
        # ping
        {},
        # diffusion.repository.search
        {"data": [{"phid": "PHID-REPO-1", "fields": {"vcs": "git"}}]},
        # user.search
        [{"userName": "alice", "phid": "PHID-USER-1"}],
        # differential.creatediff
        {"phid": "PHID-DIFF-1", "diffid": "1"},
        # differential.revision.edit
        {"object": {"id": "123", "phid": "PHID-DREV-123"}},
        # differential.setdiffproperty
        {},
    ]
    test_a = git_repo_path / "X"
    with open(test_a, "w", newline="") as f:
        f.write("a\r\nb\n")
    git_out("-c", "core.autocrlf=false", "add", "X")
    git_out("-c", "core.autocrlf=false", "commit", "-m", "A r?alice")
    # removing CR, leaving LF
    with open(test_a, "w", newline="") as f:
        f.write("a\nb\n")
    git_out("-c", "core.autocrlf=false", "add", "X")
    git_out("-c", "core.autocrlf=false", "commit", "-m", "B r?alice")

    mozphab.main(["submit", "--yes", "--bug", "1", "-s"], is_development=True)

    call_conduit.assert_any_call(
        "differential.creatediff",
        Matching(lambda x: x["changes"][0]["hunks"][0]["corpus"] == "-a\r\n+a\n b\n"),
    )


def test_submit_remove_form_feed(in_process, git_repo_path: pathlib.Path, init_sha):
    """Test deleting a file with a form feed character will not corrupt the diff."""
    call_conduit.side_effect = (
        # CREATE
        # ping
        {},
        # diffusion.repository.search
        {"data": [{"phid": "PHID-REPO-1", "fields": {"vcs": "git"}}]},
        # user.search
        [{"userName": "alice", "phid": "PHID-USER-1"}],
        # differential.creatediff
        {"phid": "PHID-DIFF-1", "diffid": "1"},
        # differential.revision.edit
        {"object": {"id": "123", "phid": "PHID-DREV-123"}},
        # differential.setdiffproperty
        {},
    )

    test_a = git_repo_path / "X"
    test_a.write_text("some line\fcontaining form feed\na second line\n")
    git_out("add", "X")
    git_out("commit", "-am", "A r?alice")

    # delete file
    test_a.unlink()
    git_out("commit", "-am", "B r?alice")

    call_conduit.reset_mock()
    mozphab.main(["submit", "--yes", "--bug", "1", "HEAD~"], is_development=True)

    failing = True
    for arg in call_conduit.call_args_list:
        if arg[0][0] == "differential.creatediff":
            assert arg[0][1]["changes"][0]["hunks"][0]["corpus"] == (
                "-some line\fcontaining form feed\n-a second line\n"
            )
            failing = False

    assert not failing


def test_submit_single_last(in_process, git_repo_path: pathlib.Path, init_sha):
    call_conduit.side_effect = (
        # ping
        {},
        # diffusion.repository.search
        {"data": [{"phid": "PHID-REPO-1", "fields": {"vcs": "git"}}]},
        # differential.creatediff
        {"phid": "PHID-DIFF-1", "diffid": "1"},
        # differential.revision.edit
        {"object": {"id": "123", "phid": "PHID-DREV-123"}},
        # differential.setdiffproperty
        {},
    )
    (git_repo_path / "X").write_text("a\n")
    git_out("add", "X")
    git_out("commit", "-am", "A")
    (git_repo_path / "X").write_text("b\n")
    git_out("commit", "-am", "B")

    mozphab.main(["submit", "--yes", "--bug", "1", "--single"], is_development=True)

    log = git_out("log", "--format=%s%n%n%b", "-2")
    expected = """\
Bug 1 - B

Differential Revision: http://example.test/D123
A


"""
    assert log == expected


def test_submit_single_first(
    in_process, git_repo_path: pathlib.Path, init_sha, git_sha
):
    call_conduit.side_effect = (
        # ping
        {},
        # diffusion.repository.search
        {"data": [{"phid": "PHID-REPO-1", "fields": {"vcs": "git"}}]},
        # differential.creatediff
        {"phid": "PHID-DIFF-1", "diffid": "1"},
        # differential.revision.edit
        {"object": {"id": "123", "phid": "PHID-DREV-123"}},
        # differential.setdiffproperty
        {},
    )
    (git_repo_path / "X").write_text("a\n")
    git_out("add", "X")
    git_out("commit", "-am", "A")
    sha = git_sha()
    (git_repo_path / "X").write_text("b\n")
    git_out("commit", "-am", "B")

    mozphab.main(
        ["submit", "--yes", "--bug", "1", "--single", sha], is_development=True
    )

    log = git_out("log", "--format=%s%n%n%b", "-2")
    expected = """\
B


Bug 1 - A

Differential Revision: http://example.test/D123
"""
    assert log == expected


def test_submit_update_no_message(in_process, git_repo_path: pathlib.Path, init_sha):
    call_conduit.reset_mock()
    call_conduit.side_effect = (
        # ping
        {},
        # diffusion.repository.search
        {"data": [{"phid": "PHID-REPO-1", "fields": {"vcs": "git"}}]},
        {"data": [search_rev(rev=123)]},
        {"data": [search_diff()]},
        {"phid": "PHID-USER-1"},
        # differential.creatediff
        {"phid": "PHID-DIFF-1", "diffid": "1"},
        # differential.revision.edit
        {"object": {"id": "123", "phid": "PHID-DREV-123"}},
        # differential.setdiffproperty
        {},
    )
    (git_repo_path / "X").write_text("ą", encoding="utf-8")
    git_out("add", ".")
    (git_repo_path / "msg").write_text(
        """\
Bug 1 - Ą

Differential Revision: http://example.test/D123
""",
        encoding="utf-8",
    )
    git_out("commit", "--file", "msg")

    mozphab.main(["submit", "--yes", "--bug", "1", init_sha], is_development=True)

    log = git_out("log", "--format=%s%n%n%b", "-1")
    expected = """\
Bug 1 - Ą

Differential Revision: http://example.test/D123

"""
    assert log == expected


def test_submit_update_revision_not_found(
    in_process, git_repo_path: pathlib.Path, init_sha, caplog: pytest.LogCaptureFixture
):
    call_conduit.reset_mock()
    call_conduit.side_effect = (
        # conduit.ping
        {},
        # diffusion.repository.search
        {"data": [{"phid": "PHID-REPO-1", "fields": {"vcs": "git"}}]},
        # differential.revision.search
        {"data": []},
    )
    testfile = git_repo_path / "X"
    testfile.write_text("ą", encoding="utf-8")
    git_out("add", "X")
    msgfile = git_repo_path / "msg"
    msgfile.write_text(
        "missing revision\n\nDifferential Revision: http://example.test/D124",
        encoding="utf-8",
    )
    git_out("commit", "--file", "msg")

    with pytest.raises(exceptions.Error, match="Unable to submit commits"):
        mozphab.main(
            ["submit", "--yes"]
            + ["--bug", "1"]
            + ["--no-wip"]
            + ["--message", "update message ćwikła"]
            + [init_sha],
            is_development=True,
        )

    assert Contains("didn't return a query result for revision D124") in caplog.messages


@mock.patch("mozphab.repository.get_lando_url_for_phabricator")
def test_uplift_create(
    m_lando_url,
    in_process,
    git_repo_path: pathlib.Path,
    init_sha: str,
    tmp_path: Path,
    caplog: pytest.LogCaptureFixture,
):
    m_lando_url.return_value = "https://lando.example.com"

    call_conduit.reset_mock()
    call_conduit.side_effect = [
        # diffusion.repository.search
        {
            "data": [
                {
                    "phid": "PHID-BETA-2",
                    "fields": {
                        "shortName": "firefox-beta",
                        "defaultBranch": "beta",
                        "callsign": "BETA",
                        "vcs": "git",
                    },
                }
            ]
        },
        # diffusion.repository.search
        {"data": [{"fields": {"shortName": "firefox-beta"}}]},
        # conduit.ping
        {},
        # differential.revision.search
        {"data": [search_rev()]},
        # differential.creatediff
        {"phid": "PHID-DIFF-2", "diffid": "2"},
        # differential.revision.edit
        {"object": {"id": "2", "phid": "PHID-DREV-2"}},
        # differential.setdiffproperty
        {},
    ]

    # Fake a `beta` branch on the `origin` remote.
    git_out("remote", "add", "origin", "https://example.com/git-repo")
    remote_branches = git_repo_path / ".git" / "refs" / "remotes" / "origin"
    remote_branches.mkdir(parents=True)
    shutil.copy(
        git_repo_path / ".git" / "refs" / "heads" / "main",
        remote_branches / "beta",
    )

    testfile = git_repo_path / "X"
    testfile.write_text("ą", encoding="utf-8")
    git_out("add", "X")
    msgfile = git_repo_path / "msg"
    msgfile.write_text(
        "Bug 1 - Ą r=alice\n\nDifferential Revision: http://example.test/D1",
        encoding="utf-8",
    )
    git_out("commit", "--file", "msg")

    out_file = tmp_path / "output.json"

    mozphab.main(
        [
            "uplift",
            "-y",
            "--train",
            "firefox-beta",
            "--output-file",
            str(out_file),
            init_sha,
        ],
        is_development=True,
    )

    call_conduit.assert_any_call(
        "differential.creatediff",
        Matching(lambda x: x.get("repositoryPHID") == "PHID-BETA-2"),
    )
    call_conduit.assert_any_call(
        "differential.revision.edit",
        {
            "transactions": [
                {"type": "title", "value": "Bug 1 - Ą"},
                {
                    "type": "summary",
                    "value": "\nOriginal Revision: http://example.test/D1",
                },
                {"type": "bugzilla.bug-id", "value": "1"},
                {"type": "update", "value": "PHID-DIFF-2"},
            ]
        },
    )
    call_conduit.assert_any_call(
        "differential.setdiffproperty",
        {
            "diff_id": "2",
            "name": "local:commits",
            "data": Contains('"summary": "Bug 1 - \\u0104"')
            & Contains(
                "Summary:\\n\\n"
                "Original Revision: http://example.test/D1\\n\\n"
                "Differential Revision: http://example.test/D2"
            ),
        },
    )
    assert (
        "Couldn't find a head for firefox-beta in version control" not in caplog.text
    ), "Git branch mapping from Phabricator was not correctly performed, and the uplift base could not be found."

    assert out_file.exists(), "Expected `--output-file` to be created."
    data = json.loads(out_file.read_text(encoding="utf-8"))

    assert "commits" in data, "Output should contain `commits` key."
    commits = data["commits"]
    assert len(commits) == 1, "One commit should have been created."
    assert data["commits"][0]["rev_id"] == "2", "Output should contain new revision."


def test_empty_file(in_process, git_repo_path: pathlib.Path, init_sha):
    # Add an empty file
    call_conduit.side_effect = (
        # ping
        {},
        # diffusion.repository.search
        {"data": [{"phid": "PHID-REPO-1", "fields": {"vcs": "git"}}]},
        # differential.creatediff
        {"phid": "PHID-DIFF-1", "diffid": "1"},
        # differential.revision.edit
        {"object": {"id": "123", "phid": "PHID-DREV-123"}},
        # differential.setdiffproperty
        {},
    )
    testfile = git_repo_path / "X"
    testfile.touch()
    git_out("add", ".")
    git_out("commit", "--message", "A")

    mozphab.main(["submit", "--yes", "--bug", "1", init_sha], is_development=True)

    log = git_out("log", "--format=%s%n%n%b", "-1")
    expected = """
Bug 1 - A

Differential Revision: http://example.test/D123
"""
    assert log.strip() == expected.strip()
    assert (
        mock.call(
            "differential.creatediff",
            {
                "changes": [
                    {
                        "metadata": {},
                        "oldPath": None,
                        "currentPath": "X",
                        "awayPaths": [],
                        "oldProperties": {},
                        "newProperties": {"unix:filemode": "100644"},
                        "commitHash": mock.ANY,
                        "type": 1,
                        "fileType": 1,
                        "hunks": [],
                    }
                ],
                "sourceMachine": "http://example.test",
                "sourceControlSystem": "git",
                "sourceControlPath": "/",
                "sourceControlBaseRevision": mock.ANY,
                "creationMethod": "moz-phab-git",
                "lintStatus": "none",
                "unitStatus": "none",
                "repositoryPHID": "PHID-REPO-1",
                "sourcePath": mock.ANY,
                "branch": "HEAD",
            },
        )
        in call_conduit.call_args_list
    ), (
        "The diff should have populated fields and "
        "`change.hunks` should be an empty list"
    )

    # Modify a file's mode
    call_conduit.reset_mock()
    call_conduit.side_effect = (
        # differential.creatediff
        {"phid": "PHID-DIFF-2", "diffid": "2"},
        # differential.revision.edit
        {"object": {"id": "124", "phid": "PHID-DREV-124"}},
        # differential.setdiffproperty
        {},
    )
    # Windows cannot update UMASK bits in the same way Unix
    # systems can, so use Git's update-index command to
    # directly modify the index for the file.
    if platform.system() == "Windows":
        git_out("update-index", "--chmod=+x", str(testfile))
    else:
        os.chmod(testfile, 0o0755)
    git_out("commit", "-a", "--message", "B")
    mozphab.main(["submit", "--yes", "--bug", "1", "HEAD~"], is_development=True)
    log = git_out("log", "--format=%s%n%n%b", "-1")
    expected = """
Bug 1 - B

Differential Revision: http://example.test/D124
"""
    assert log.strip() == expected.strip()
    assert (
        mock.call(
            "differential.creatediff",
            {
                "changes": [
                    {
                        "metadata": {},
                        "oldPath": "X",
                        "currentPath": "X",
                        "awayPaths": [],
                        "oldProperties": {"unix:filemode": "100644"},
                        "newProperties": {"unix:filemode": "100755"},
                        "commitHash": mock.ANY,
                        "type": 2,
                        "fileType": 1,
                        "hunks": [],
                    }
                ],
                "sourceMachine": "http://example.test",
                "sourceControlSystem": "git",
                "sourceControlPath": "/",
                "sourceControlBaseRevision": mock.ANY,
                "creationMethod": "moz-phab-git",
                "lintStatus": "none",
                "unitStatus": "none",
                "repositoryPHID": "PHID-REPO-1",
                "sourcePath": mock.ANY,
                "branch": "HEAD",
            },
        )
        in call_conduit.call_args_list
    ), (
        "The diff should contain filemode changes and "
        "`change.hunks` should be an empty list"
    )

    # Remove an empty file
    call_conduit.reset_mock()
    call_conduit.side_effect = (
        # differential.creatediff
        {"phid": "PHID-DIFF-3", "diffid": "3"},
        # differential.revision.edit
        {"object": {"id": "125", "phid": "PHID-DREV-125"}},
        # differential.setdiffproperty
        {},
    )
    testfile.unlink()
    git_out("commit", "-a", "--message", "C")
    mozphab.main(["submit", "--yes", "--bug", "1", "HEAD~"], is_development=True)
    log = git_out("log", "--format=%s%n%n%b", "-1")
    expected = """
Bug 1 - C

Differential Revision: http://example.test/D125
"""
    assert log.strip() == expected.strip()
    assert (
        mock.call(
            "differential.creatediff",
            {
                "changes": [
                    {
                        "metadata": {},
                        "oldPath": "X",
                        "currentPath": "X",
                        "awayPaths": [],
                        "oldProperties": {"unix:filemode": "100755"},
                        "newProperties": {},
                        "commitHash": mock.ANY,
                        "type": 3,
                        "fileType": 1,
                        "hunks": [],
                    }
                ],
                "sourceMachine": "http://example.test",
                "sourceControlSystem": "git",
                "sourceControlPath": "/",
                "sourceControlBaseRevision": mock.ANY,
                "creationMethod": "moz-phab-git",
                "lintStatus": "none",
                "unitStatus": "none",
                "repositoryPHID": "PHID-REPO-1",
                "sourcePath": mock.ANY,
                "branch": "HEAD",
            },
        )
        in call_conduit.call_args_list
    ), (
        "The diff should have empty `newProperties` and "
        "`change.hunks` should be an empty list"
    )



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import argparse
import re
from typing import List

from mozphab.conduit import conduit
from mozphab.exceptions import Error, NotFoundError
from mozphab.helpers import PHABRICATOR_URL_REVISION_PATTERN, prompt
from mozphab.logger import logger
from mozphab.spinner import wait_message


def check_revision_id(value: str) -> int:
    """Parse the revision ID from `value`.

    `value` is a `str` which is either `<id>`, `D<id>`,
    or a Phabricator revision URL.
    """
    # D123 or 123
    m = re.search(r"^D?(\d+)$", value)
    if m:
        return int(m.group(1))

    # Full URL
    m = re.search(r"^" + PHABRICATOR_URL_REVISION_PATTERN, value)
    if m:
        return int(m.group("rev"))

    # Invalid
    raise argparse.ArgumentTypeError(
        "Invalid Revision ID (expected number or URL): %s\n" % value
    )


def abandon_revisions(revision_ids: List[int], args: argparse.Namespace):
    """Abandon the specified revisions in Phabricator."""

    # Check connection to Phabricator
    with wait_message("Checking connection to Phabricator..."):
        if not conduit.check():
            raise Error("Failed to use Conduit API")

    # Get revision information
    with wait_message("Fetching revision information..."):
        try:
            revisions = conduit.get_revisions(ids=revision_ids)
        except NotFoundError as e:
            raise Error(f"Failed to fetch revisions: {e}")

    if not revisions:
        raise Error("No revisions found")

    # Filter out revisions that are already abandoned
    revisions_to_abandon = []
    already_abandoned = []

    for revision in revisions:
        status = revision["fields"]["status"]["value"]
        if status == "abandoned":
            already_abandoned.append(revision["id"])
        else:
            revisions_to_abandon.append(revision)

    # Show what will be abandoned
    if already_abandoned:
        logger.info(
            "Already abandoned: %s",
            ", ".join(f"D{rev_id}" for rev_id in already_abandoned),
        )

    if not revisions_to_abandon:
        logger.warning("All specified revisions are already abandoned.")
        return

    logger.warning("The following revisions will be abandoned:")
    for revision in revisions_to_abandon:
        logger.info(" * D%s: %s", revision["id"], revision["fields"]["title"])

    # Confirmation prompt
    if not args.yes:
        res = prompt("Abandon these revisions?", ["Yes", "No"])
        if res == "No":
            return

    # Abandon the revisions
    with wait_message("Abandoning revisions..."):
        for revision in revisions_to_abandon:
            transactions = [{"type": "abandon", "value": True}]
            conduit.apply_transactions_to_revision(
                rev_id=revision["phid"], transactions=transactions
            )
            logger.info("D%s abandoned", revision["id"])

    logger.warning("Completed: %d revision(s) abandoned", len(revisions_to_abandon))


def abandon(repo, args: argparse.Namespace):
    """Main function for the abandon command."""
    abandon_revisions(args.revisions, args)


def add_parser(parser):
    """Add the abandon command parser."""
    abandon_parser = parser.add_parser(
        "abandon", help="Abandon revisions in Phabricator."
    )
    abandon_parser.add_argument(
        "revisions",
        nargs="+",
        type=check_revision_id,
        help="Revision(s) to abandon. Supports IDs (123, D123) and URLs.",
    )
    abandon_parser.add_argument(
        "--yes",
        "-y",
        action="store_true",
        help="Abandon without confirmation (default: False).",
    )
    abandon_parser.add_argument(
        "--safe-mode",
        dest="safe_mode",
        action="store_true",
        help="Run VCS with only necessary extensions.",
    )
    abandon_parser.set_defaults(func=abandon, needs_repo=True)



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import logging
from pathlib import Path
from unittest import mock

import pytest
from callee import Contains

from mozphab import environment, exceptions, mozphab
from mozphab.commits import Commit

from .conftest import create_temp_fn


@pytest.mark.parametrize(
    "revisions, expected",
    (
        (
            [
                "a1b2c3d4e5f6g7h8i9j0aabbccddeeff00112233",  # oldest unpublished
                "b1c2d3e4f5g6h7i8j9k0aabbccddeeff00112234",
                "c1d2e3f4g5h6i7j8k9l0aabbccddeeff00112235",  # newest unpublished
                "-1234567890abcdef1234567890abcdef12345678",  # last published commit
            ],
            "c1d2e3f4g5h6i7j8k9l0aabbccddeeff00112235",
        ),
        (["-abcdef1234567890abcdef1234567890abcdef12"], None),
        ([], None),
    ),
)
@mock.patch("mozphab.git.Git.git_out")
def test_first_unpublished(m_git_git_out, git, revisions, expected):
    m_git_git_out.return_value = revisions

    result = git._get_first_unpublished_node()
    assert result == expected, "Incorrect first unpublished node detected."


@mock.patch("mozphab.git.config")
@mock.patch("mozphab.git.Git.git_out")
def test_get_base_remotes_with_git_remote(mock_git_out, mock_config, git):
    mock_config.git_remote = ["custom"]
    mock_git_out.return_value = ["origin", "upstream"]

    remotes = git.get_base_remotes()
    assert remotes == [
        "custom"
    ], "`git_remote` config option should override any found remotes."
    mock_git_out.assert_not_called()


@mock.patch("mozphab.git.config")
@mock.patch("mozphab.git.Git.git_out")
def test_get_base_remotes_with_remote_arg(mock_git_out, mock_config, git):
    mock_config.git_remote = ["origin"]
    git.args.upstream = ["custom"]

    remotes = git.get_base_remotes()
    assert remotes == ["custom"], "--upstream arg should take priority."
    mock_git_out.assert_not_called()


@mock.patch("mozphab.git.config")
@mock.patch("mozphab.git.Git.git_out")
def test_get_base_remotes_single(
    mock_git_out, mock_config, git, caplog: pytest.LogCaptureFixture
):
    caplog.set_level(logging.INFO)
    mock_config.git_remote = []
    mock_git_out.return_value = ["onlyremote"]

    remotes = git.get_base_remotes()

    assert remotes == [
        "onlyremote"
    ], "Single remote in the repo should be returned as the base."
    assert caplog.messages == ["Using the only available remote: onlyremote"]


@mock.patch("mozphab.git.config")
@mock.patch("mozphab.git.Git.git_out")
def test_get_base_remotes_with_origin(
    mock_git_out, mock_config, git, caplog: pytest.LogCaptureFixture
):
    mock_config.git_remote = []
    mock_git_out.return_value = ["upstream", "origin", "custom"]

    remotes = git.get_base_remotes()

    assert remotes == [
        "origin"
    ], "`origin` should be returned when multiple remotes are found."
    assert caplog.messages == [Contains("Defaulting to 'origin'")]


@mock.patch("mozphab.git.config")
@mock.patch("mozphab.git.Git.git_out")
def test_get_base_remotes_multiple_without_origin(
    mock_git_out, mock_config, git, caplog: pytest.LogCaptureFixture
):
    mock_config.git_remote = []
    mock_git_out.return_value = ["upstream", "custom"]

    remotes = git.get_base_remotes()

    assert remotes == [
        "upstream",
        "custom",
    ], "All remotes should be returned when multiple remotes found without `origin`."
    assert caplog.messages == [
        Contains("Multiple remotes found, and no `origin` present.")
    ]


@mock.patch("mozphab.git.Git.get_base_remotes")
def test_get_base_remote_args_with_result(mock_get_base_remotes, git):
    mock_get_base_remotes.return_value = ["origin", "custom"]
    remote_args = git.get_base_remote_args()
    assert remote_args == ["--remotes=origin", "--remotes=custom"]


@mock.patch("mozphab.git.Git.get_base_remotes")
def test_get_base_remote_args_empty(mock_get_base_remotes, git):
    mock_get_base_remotes.return_value = []
    remote_args = git.get_base_remote_args()
    assert remote_args == ["--remotes"]


@mock.patch("mozphab.git.Git.git_out")
def test_branches_to_rebase(m_git_git_out, git):
    git_find = git._find_branches_to_rebase

    # No branch returned - not a real case - we don't work without branches
    m_git_git_out.return_value = []
    assert {} == git_find([Commit(orig_node="_aaa", node="aaa")])

    # No amend, no branches to rebase
    m_git_git_out.return_value = ["branch"]
    assert {} == git_find([Commit(orig_node="aaa", node="aaa")])

    # One commit, one branch
    m_git_git_out.return_value = ["branch"]
    assert {"branch": ["aaa", "_aaa"]} == git_find(
        [Commit(orig_node="_aaa", node="aaa")]
    )

    # Two commits one branch
    m_git_git_out.return_value = ["branch"]
    assert {"branch": ["bbb", "_bbb"]} == git_find(
        [Commit(orig_node="_aaa", node="aaa"), Commit(orig_node="_bbb", node="bbb")]
    )

    # Two branches one commit
    # ... (branch1)
    # | ... (branch2)
    # |/
    # * aaa
    # More realistic output from the git command
    m_git_git_out.return_value = ["*  branch1", "  branch2"]
    assert {"branch1": ["aaa", "_aaa"], "branch2": ["aaa", "_aaa"]} == git_find(
        [Commit(orig_node="_aaa", node="aaa")]
    )

    # ... (branch1)
    # | * bbb (branch2)
    # |/
    # * aaa
    m_git_git_out.side_effect = (["branch1", "branch2"], ["branch2"])
    assert {"branch1": ["aaa", "_aaa"], "branch2": ["bbb", "_bbb"]} == git_find(
        [Commit(orig_node="_aaa", node="aaa"), Commit(orig_node="_bbb", node="bbb")]
    )

    # * ... (master)
    # | * ... (feature1)
    # | | * ... (feature2)
    # | |/
    # |/|
    # | | * ddd (feature1_1)
    # | |/
    # | * ccc
    # |/
    # * bbb
    # * aaa

    m_git_git_out.side_effect = (
        ["master", "feature1", "feature1_1", "feature2"],  # aaa
        ["master", "feature1", "feature1_1", "feature2"],  # bbb
        ["feature1", "feature1_1"],  # ccc
        ["feature1_1"],  # ddd
    )
    assert {
        "master": ["bbb", "_bbb"],
        "feature1": ["ccc", "_ccc"],
        "feature2": ["bbb", "_bbb"],
        "feature1_1": ["ddd", "_ddd"],
    } == git_find(
        [
            Commit(orig_node="_aaa", node="aaa"),
            Commit(orig_node="_bbb", node="bbb"),
            Commit(orig_node="_ccc", node="ccc"),
            Commit(orig_node="_ddd", node="ddd"),
        ]
    )


def test_get_direct_children(git):
    get_children = git._get_direct_children
    rev_list = ["aaa bbb ccc", "bbb", "ccc ddd"]
    assert ["bbb", "ccc"] == get_children("aaa", rev_list)
    assert [] == get_children("bbb", rev_list)
    assert ["ddd"] == get_children("ccc", rev_list)
    assert [] == get_children("xxx", rev_list)


def test_is_child(git):
    is_child = git._is_child
    # * ccc
    # * bbb
    # * aaa
    nodes = ["ccc", "bbb ccc", "aaa bbb"]
    assert is_child("aaa", "bbb", nodes)
    assert is_child("aaa", "ccc", nodes)
    assert is_child("bbb", "ccc", nodes)
    assert not is_child("bbb", "aaa", nodes)
    assert not is_child("aaa", "aaa", nodes)
    assert not is_child("bbb", "bbb", nodes)
    assert not is_child("ccc", "ccc", nodes)

    # * ddd
    # | * ccc
    # | | * eee
    # | |/
    # | * bbb
    # |/
    # * aaa
    nodes = ["ddd", "ccc", "eee", "bbb ccc eee", "aaa bbb ddd"]
    assert is_child("aaa", "bbb", nodes)
    assert is_child("aaa", "ccc", nodes)
    assert is_child("aaa", "ddd", nodes)
    assert is_child("aaa", "eee", nodes)
    assert is_child("bbb", "ccc", nodes)
    assert is_child("bbb", "eee", nodes)
    assert not is_child("bbb", "ddd", nodes)
    assert not is_child("ccc", "ddd", nodes)


@mock.patch("mozphab.git.Git.git_out")
@mock.patch("mozphab.mozphab.config")
def test_range(m_config, m_git_git_out, git):
    class Args:
        def __init__(self, start="start", end="."):
            self.start_rev = start
            self.end_rev = end
            self.safe_mode = False
            self.single = False

    m_config.safe_mode = False
    m_git_git_out.return_value = ["user.email=email"]
    git.set_args(Args())
    assert git.revset == ("start", ".")


@mock.patch("mozphab.mozphab.config")
@mock.patch("mozphab.gitcommand.parse_config")
@mock.patch("mozphab.git.Git._get_first_unpublished_node")
@mock.patch("mozphab.git.Git.git_out")
def test_set_args(m_git_git_out, m_git_get_first, m_parse_config, m_config, git):
    class Args:
        def __init__(
            self,
            start=environment.DEFAULT_START_REV,
            end=environment.DEFAULT_END_REV,
            safe_mode=False,
            single=False,
        ):
            self.start_rev = start
            self.end_rev = end
            self.safe_mode = safe_mode
            self.single = single

    with pytest.raises(exceptions.Error):
        git.set_args(Args())

    git.git.command = ["git"]
    git.git.safe_mode = False
    m_parse_config.return_value = {"user.email": "email"}
    m_git_get_first.return_value = "aaa"
    git.set_args(Args())
    assert ["git"] == git.git.command
    m_git_get_first.assert_called_once()
    assert git.revset == ("aaa^", ".")

    m_parse_config.return_value = {
        "user.email": "email",
        "user.name": "name",
        "cinnabar.helper": "string",
    }
    git.set_args(Args())
    assert ["git"] == git.git.command
    assert ["cinnabar"] == git.git.extensions

    safe_options = (
        ["git"]
        + ["-c", "user.email=email"]
        + ["-c", "user.name=name"]
        + ["-c", "cinnabar.helper=string"]
    )
    git.set_args(Args(safe_mode=True))
    assert safe_options == git.git.command

    git.git.command = ["git"]
    git.git.safe_mode = True
    git.set_args(Args())
    assert safe_options == git.git.command

    m_config.safe_mode = False
    m_git_get_first.reset_mock()
    git.set_args(Args(start="bbb", end="ccc"))
    m_git_get_first.assert_not_called()
    assert git.revset == ("bbb", "ccc")

    git.set_args(Args(safe_mode=True))
    assert "" == git.git._env["HOME"]
    assert "" == git.git._env["XDG_CONFIG_HOME"]

    m_config.safe_mode = True
    git.set_args(Args())
    assert "" == git.git._env["HOME"]
    assert "" == git.git._env["XDG_CONFIG_HOME"]

    m_git_get_first.reset_mock()
    m_parse_config.return_value = {
        "user.email": "email",
        "user.name": "name",
    }
    git.set_args(Args(single=True))
    m_git_get_first.assert_not_called()
    assert git.revset == ("HEAD^", "HEAD")

    git.set_args(Args(single=True, start="start"))
    m_git_get_first.assert_not_called()
    assert git.revset == ("start^", "start")


@mock.patch("mozphab.git.Git.git_out")
def test_worktree_clean(m_git_out, git):
    m_git_out.return_value = ""
    assert git.is_worktree_clean()

    m_git_out.return_value = ["xxx"]
    assert not git.is_worktree_clean()

    m_git_out.return_value = ["?? one", "?? two"]
    assert git.is_worktree_clean()

    m_git_out.return_value = ["?? one", "?? two", " M xxx"]
    assert not git.is_worktree_clean()


@mock.patch("mozphab.git.Git.git_call")
def test_commit(m_git, git):
    git.commit("some body")
    m_git.assert_called_once()

    m_git.reset_mock()
    git.commit("some body", "user")
    m_git.assert_called_once()


@mock.patch("mozphab.git.Git._hg_to_git")
@mock.patch("mozphab.git.Git.is_node")
@mock.patch("mozphab.git.Git.phab_vcs")
def test_check_node(m_phab_vcs, m_git_is_node, m_hg2git, git):
    node = "aabbcc"
    mozphab.conduit.set_repo(git)
    assert node == git.check_node(node)

    git._phab_vcs = "hg"
    git.git._cinnabar_installed = False
    m_git_is_node.return_value = False
    with pytest.raises(exceptions.NotFoundError) as e:
        git.check_node(node)
    assert "Cinnabar extension not enabled" in str(e.value)

    git.git._cinnabar_installed = True
    m_hg2git.return_value = "0" * 40
    with pytest.raises(exceptions.NotFoundError) as e:
        git.check_node(node)
    assert "Mercurial SHA1 not found" in str(e.value)

    m_hg2git.return_value = "git_aabbcc"
    with pytest.raises(exceptions.NotFoundError) as e:
        git.check_node(node)
    assert "Mercurial SHA1 detected, but commit not found" in str(e.value)

    m_git_is_node.side_effect = (False, True)
    assert "git_aabbcc" == git.check_node(node)


@mock.patch("mozphab.git.Git.git_out")
@mock.patch("mozphab.git.Git.checkout")
@mock.patch("mozphab.git.Git.git_call")
@mock.patch("mozphab.git.prompt")
@mock.patch("mozphab.git.config")
def test_before_patch(
    m_config,
    m_prompt,
    m_git,
    m_checkout,
    m_git_out,
    git,
    caplog: pytest.LogCaptureFixture,
):
    class Args:
        def __init__(
            self,
            rev_id="D123",
            nocommit=False,
            raw=False,
            applyto="base",
            no_branch=False,
            yes=False,
        ):
            self.rev_id = rev_id
            self.nocommit = nocommit
            self.raw = raw
            self.applyto = applyto
            self.no_branch = no_branch
            self.yes = yes

    git.args = Args()
    m_git_out.side_effect = (["  branch"],)
    git.before_patch("sha1", "branch")
    m_checkout.assert_called_with("sha1")
    m_git.assert_called_with(["checkout", "-q", "-b", "branch_1"])

    m_git.reset_mock()
    m_git_out.reset_mock()
    m_checkout.reset_mock()

    m_checkout.reset_mock()
    m_git_out.side_effect = ("the branch name is not here",)
    git.args = Args(applyto="here")
    git.before_patch(None, "branchname")
    m_checkout.assert_not_called()

    m_git.reset_mock()
    m_checkout.reset_mock()
    git.args = Args(applyto="abcdef", nocommit=True)
    git.before_patch("abcdef", None)
    m_checkout.assert_called_once_with("abcdef")
    m_git.assert_not_called()

    m_git.reset_mock()
    m_checkout.reset_mock()
    caplog.clear()
    git.args = Args(no_branch=True, yes=True)
    git.before_patch("abcdef", "name")
    m_checkout.assert_called_once()
    m_git.assert_not_called()
    assert Contains("git checkout -b") in caplog.messages

    m_git.reset_mock()
    m_checkout.reset_mock()
    caplog.clear()
    git.args = Args(no_branch=True)
    git.before_patch("abcdef", "name")
    m_checkout.assert_called_once()
    m_git.assert_not_called()
    m_prompt.assert_called_once()
    assert Contains("git checkout -b") in caplog.messages

    m_git.reset_mock()
    m_checkout.reset_mock()
    caplog.clear()
    m_prompt.reset_mock()
    m_config.create_branch = False
    git.args = Args()
    git.before_patch("abcdef", "name")
    m_checkout.assert_called_once()
    m_git.assert_not_called()
    m_prompt.assert_called_once()
    assert Contains("git checkout -b") in caplog.messages

    m_prompt.return_value = "No"
    with pytest.raises(SystemExit):
        git.before_patch("abcdef", "name")


@mock.patch("mozphab.git.temporary_binary_file")
@mock.patch("mozphab.git.Git.git_call")
@mock.patch("mozphab.git.Git.commit")
def test_apply_patch(m_commit, m_git, m_temp_fn, git):
    m_temp_fn.return_value = create_temp_fn("filename")
    git.apply_patch("diff", "commit message", "user", 1)
    m_git.assert_called_once_with(["apply", "--index", "filename"])
    m_commit.assert_called_with("commit message", "user", 1)
    m_temp_fn.assert_called_once_with(b"diff")


@mock.patch("mozphab.git.Git.git_out")
def test_is_node(m_git_out, git):
    m_git_out.return_value = "commit"
    assert git.is_node("aaa")

    m_git_out.return_value = "something else"
    assert not git.is_node("aaa")

    m_git_out.side_effect = exceptions.CommandError
    assert not git.is_node("aaa")


@mock.patch("mozphab.git.Git.git_out")
def test_is_descendant(m_git_out, git):
    git.revset = ("abc", "def")

    m_git_out.return_value = ""
    assert (
        git.is_descendant("aabbcc") is True
    ), "Call to merge-base with 0 return code should indicate node is a descendant."

    m_git_out.side_effect = exceptions.CommandError("", 1)
    assert (
        git.is_descendant("aabbcc") is False
    ), "Call to merge-base with 1 return code should indicated node is not descendant."

    m_git_out.side_effect = exceptions.CommandError("test", 255)
    with pytest.raises(exceptions.CommandError) as e:
        git.is_descendant("aabbcc")

        assert (
            e.args == "test"
        ), "Original command exception was not raised to the caller."


@mock.patch("mozphab.gitcommand.which")
@mock.patch("mozphab.gitcommand.GitCommand.output")
def test_is_cinnabar_installed(m_git_out, m_which, git, tmp_path):
    def _without_str(calls):
        # Debuggers call __str__ on mocked functions, strip them
        return [c for c in calls if c[0] != "__str__"]

    # cinnabar installed as visible external command
    m_git_out.return_value = ["External commands", "   cinnabar", "   revise"]
    assert git.is_cinnabar_installed
    m_git_out.assert_called_once_with(["help", "--all"])

    # cached request (primed by 'cinnabar installed' test)
    m_git_out.reset_mock()
    m_git_out.return_value = ["External commands", "   cinnabar", "   revise"]
    assert git.is_cinnabar_installed
    m_git_out.assert_not_called()

    # create a fake cinnabar in exec-path for git to find
    cinnabar = Path(tmp_path) / "git-cinnabar"
    cinnabar.write_text("")
    cinnabar.chmod(0o755)

    # cinnabar installed in exec-path
    m_git_out.reset_mock()
    m_git_out.side_effect = [
        ["External commands", "   revise"],  # git help --all
        tmp_path,  # git --exec-path
    ]
    git.git._cinnabar_installed = None
    assert git.is_cinnabar_installed
    assert _without_str(m_git_out.mock_calls) == [
        mock.call(["help", "--all"]),
        mock.call(["--exec-path"], split=False),
    ]

    # remove cinnabar from exec-path so we fall back to looking on the path
    cinnabar.unlink()

    # cinnabar installed somewhere on path
    m_git_out.reset_mock()
    m_git_out.side_effect = [
        ["External commands", "   revise"],  # git help --all
        tmp_path,  # git --exec-path
    ]
    m_which.return_value = str(cinnabar)
    git.git._cinnabar_installed = None
    assert git.is_cinnabar_installed
    assert _without_str(m_git_out.mock_calls) == [
        mock.call(["help", "--all"]),
        mock.call(["--exec-path"], split=False),
    ]

    # cinnabar not installed
    m_git_out.reset_mock()
    m_git_out.side_effect = [
        ["External commands", "   revise"],  # git help --all
        tmp_path,  # git --exec-path
    ]
    m_which.return_value = None
    git.git._cinnabar_installed = None
    assert not git.is_cinnabar_installed
    assert _without_str(m_git_out.mock_calls) == [
        mock.call(["help", "--all"]),
        mock.call(["--exec-path"], split=False),
    ]


@mock.patch("mozphab.git.Git.git_out")
def test_unicode_in_windows_env(m_git_out, git, monkeypatch):
    monkeypatch.setattr(environment, "IS_WINDOWS", True)
    git._commit_tree("parent", "tree_hash", "message", "ćwikła", "ćwikła", "date")
    m_git_out.assert_called_once_with(
        ["commit-tree", "-p", "parent", "-F", mock.ANY, "tree_hash"],
        split=False,
        extra_env={
            "GIT_AUTHOR_NAME": "ćwikła",
            "GIT_AUTHOR_EMAIL": "ćwikła",
            "GIT_AUTHOR_DATE": "date",
        },
    )


def test_check_vcs(git):
    class Args:
        def __init__(self, force_vcs=False):
            self.safe_mode = False
            self.force_vcs = force_vcs

    args = Args()
    git.set_args(args)
    assert git.check_vcs()

    git.git._cinnabar_installed = True
    git._phab_vcs = "hg"
    assert git.check_vcs()

    git.git._cinnabar_installed = False

    # This just produces a warning now.
    # See bug 1963878.
    git.check_vcs()

    args = Args(force_vcs=True)
    git.set_args(args)
    assert git.check_vcs()


@mock.patch("mozphab.git.Git._get_commits_info")
@mock.patch("mozphab.git.Git._git_get_children")
@mock.patch("mozphab.git.Git._is_child")
def test_commit_stack_single(_1, _2, _3, git):
    git._get_commits_info.return_value = [
        """\
Tue, 22 Jan 2019 13:42:48 +0000
Conduit User
conduit@mozilla.bugs
4912923
b18312ffe929d3482f1d7b1e9716a1885c7a61b8
aaa000aaa000aaa000aaa000aaa000aaa000aaa0
title

description
"""
    ]
    git.revset = ["HEAD^", "HEAD"]
    git.commit_stack(single=True)
    git._git_get_children.assert_not_called()
    git._is_child.assert_not_called()


@pytest.mark.parametrize(
    "cinnabar_required,remotes,unified_head",
    (
        (True, ["origin"], "remotes/origin/bookmarks/beta"),
        (False, ["origin"], "remotes/origin/beta"),
        (True, ["origin", "upstream"], "remotes/origin/bookmarks/beta"),
        (False, ["origin", "upstream"], "remotes/origin/beta"),
        (True, ["upstream"], "remotes/upstream/bookmarks/beta"),
        (False, ["upstream"], "remotes/upstream/beta"),
    ),
)
@mock.patch("mozphab.git.Git.is_node")
@mock.patch("mozphab.git.Git.is_cinnabar_required", new_callable=mock.PropertyMock)
@mock.patch("mozphab.git.Git.get_base_remotes")
def test_git_get_repo_head_branch(
    m_get_base_remotes,
    m_is_cinnabar_required,
    m_is_node,
    git,
    cinnabar_required,
    remotes,
    unified_head,
):
    m_get_base_remotes.return_value = remotes
    m_is_cinnabar_required.return_value = cinnabar_required

    # If head is not a node in the repo, raise `ValueError`.
    m_is_node.return_value = False
    assert (
        git.get_repo_head_branch() is None
    ), "No found heads should have returned `None`."

    # If head is a node in the repo, should map to a remote branch.
    m_is_node.return_value = True
    assert (
        git.get_repo_head_branch() == unified_head
    ), "beta did not correctly map to a branch"


def test_git_validate_email(git):
    with pytest.raises(exceptions.Error):
        git.validate_email()

    git.git.email = "test@mozilla.com"
    assert git.validate_email() is None, "validate_email() executes without error"



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from unittest import mock

from mozphab import mozphab

from .conftest import hg_out

_revision = 100

call_conduit = mock.Mock()


def _init_repo(hg_repo_path):
    test_file = hg_repo_path / "X"
    test_file.write_text("R0")
    hg_out("commit", "--addremove", "--message", "R0")
    return {"test_file": test_file, "rev": 1, "rev_map": {"R0": "1"}}


def _add_commit(repo, parent, name):
    hg_out("update", repo["rev_map"][parent])
    repo["test_file"].write_text(name)
    hg_out("commit", "--message", name)
    repo["rev"] += 1
    repo["rev_map"][name] = str(repo["rev"])


def _checkout(repo, name):
    hg_out("update", repo["rev_map"][name])


def _submit(repo, start, end, expected, wip=False):
    mozphab.main(
        [
            "submit",
            "--yes",
            *(["--wip"] if wip else []),
            "--bug",
            "1",
            repo["rev_map"][start],
            repo["rev_map"][end],
        ],
        is_development=True,
    )
    log = hg_out("log", "--graph", "--template", r"{desc|firstline}\n")
    assert log.strip() == expected.strip()


def _conduit_side_effect(calls=1):
    side_effect = [
        # ping
        {},
        # diffusion.repository.search
        {"data": [{"phid": "PHID-REPO-1", "fields": {"vcs": "hg"}}]},
    ]

    for i in range(calls):
        side_effect.extend(
            [
                # differential.creatediff
                {"phid": "PHID-DIFF-{}".format(i), "diffid": str(i)},
                # differential.revision.edit
                {"object": {"id": str(123 + i), "phid": f"PHID-DREV-{str(123 + 1)}"}},
                # differential.setdiffproperty
                {},
            ]
        )

    return side_effect


def test_submit_wip(in_process, hg_repo_path):
    repo = _init_repo(hg_repo_path)

    call_conduit.side_effect = _conduit_side_effect()  # TODO verify bool

    _add_commit(repo, "R0", "A1")
    _submit(
        repo,
        "A1",
        "A1",
        """
@  Bug 1 - A1
|
o  R0
|
o  init
""",
        wip=True,
    )
    # Get the API call arguments for the `differential.revision.edit` call.
    edit_call_api_call_args = next(
        args[1]
        for _name, args, _kwargs in call_conduit.mock_calls
        if args[0] == "differential.revision.edit"
    )
    # Get the value of the `plan-changes` transaction.
    wip_transaction = next(
        transaction
        for transaction in edit_call_api_call_args["transactions"]
        if transaction["type"] == "plan-changes"
    )
    assert (
        wip_transaction["value"] is True
    ), "`wip=True` in `submit` should set `plan-changes`."


def test_submit_single_1(in_process, hg_repo_path):
    repo = _init_repo(hg_repo_path)

    call_conduit.side_effect = _conduit_side_effect()

    _add_commit(repo, "R0", "A1")
    _add_commit(repo, "A1", "B1")
    _add_commit(repo, "B1", "C1")
    _submit(
        repo,
        "A1",
        "A1",
        """
@  C1
|
o  B1
|
o  Bug 1 - A1
|
o  R0
|
o  init
""",
    )


def test_submit_single_2(in_process, hg_repo_path):
    repo = _init_repo(hg_repo_path)

    call_conduit.side_effect = _conduit_side_effect()

    _add_commit(repo, "R0", "A1")
    _add_commit(repo, "A1", "B1")
    _add_commit(repo, "A1", "B2")
    _submit(
        repo,
        "A1",
        "A1",
        """
@  B2
|
| o  B1
|/
o  Bug 1 - A1
|
o  R0
|
o  init
""",
    )


def test_submit_single_3(in_process, hg_repo_path):
    repo = _init_repo(hg_repo_path)

    call_conduit.side_effect = _conduit_side_effect()

    _add_commit(repo, "R0", "A1")
    _add_commit(repo, "A1", "B1")
    _add_commit(repo, "B1", "C1")
    _add_commit(repo, "B1", "C2")
    _submit(
        repo,
        "A1",
        "A1",
        """
@  C2
|
| o  C1
|/
o  B1
|
o  Bug 1 - A1
|
o  R0
|
o  init
""",
    )


def test_submit_stack_1(in_process, hg_repo_path):
    repo = _init_repo(hg_repo_path)

    call_conduit.side_effect = _conduit_side_effect(2)

    _add_commit(repo, "R0", "A1")
    _add_commit(repo, "A1", "B1")
    _checkout(repo, "A1")
    _submit(
        repo,
        "A1",
        "B1",
        """
o  Bug 1 - B1
|
@  Bug 1 - A1
|
o  R0
|
o  init
""",
    )


def test_submit_stack_2(in_process, hg_repo_path):
    repo = _init_repo(hg_repo_path)

    call_conduit.side_effect = _conduit_side_effect(2)

    _add_commit(repo, "R0", "A1")
    _add_commit(repo, "A1", "B1")
    _add_commit(repo, "A1", "B2")
    _submit(
        repo,
        "A1",
        "B1",
        """
o  Bug 1 - B1
|
| @  B2
|/
o  Bug 1 - A1
|
o  R0
|
o  init
""",
    )


def test_submit_stack_3(in_process, hg_repo_path):
    repo = _init_repo(hg_repo_path)

    call_conduit.side_effect = _conduit_side_effect(2)

    _add_commit(repo, "R0", "A1")
    _add_commit(repo, "A1", "B1")
    _add_commit(repo, "A1", "B2")
    _add_commit(repo, "B1", "C1")
    _submit(
        repo,
        "A1",
        "B1",
        """
@  C1
|
o  Bug 1 - B1
|
| o  B2
|/
o  Bug 1 - A1
|
o  R0
|
o  init
""",
    )



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import json

from mozphab.commits import Commit


def test_commit_to_dict_json_roundtrip():
    commit = Commit(
        name="commit-1",
        node="abcdef123456",
        orig_node="fedcba654321",
        submit=True,
        title="Add feature X",
        title_preview="Add feature X (preview)",
        body="Implements feature X and adds tests.",
        author_date_epoch=1727200000,
        author_name="Author Name",
        author_email="author@example.com",
        author_date="2024-09-25 12:34:56",
        parent="123456abcdef",
        bug_id="190001",
        bug_id_orig="190001",
        rev_id=12345,
        rev_phid="PHID-DREV-xyz",
        wip=False,
        tree_hash="deadbeefcafebabe",
        reviewers={"granted": ["r1", "r2"], "request": ["r3"]},
    )

    # Convert to dict, then JSON, then back to dict.
    commit_dict = commit.to_dict()
    json_blob = json.dumps(commit_dict)  # should not raise
    commit_dict_loaded = json.loads(json_blob)

    expected = {
        "name": "commit-1",
        "node": "abcdef123456",
        "orig_node": "fedcba654321",
        "submit": True,
        "title": "Add feature X",
        "title_preview": "Add feature X (preview)",
        "body": "Implements feature X and adds tests.",
        "author_date_epoch": 1727200000,
        "author_name": "Author Name",
        "author_email": "author@example.com",
        "author_date": "2024-09-25 12:34:56",
        "parent": "123456abcdef",
        "bug_id": "190001",
        "bug_id_orig": "190001",
        "rev_id": 12345,
        "rev_phid": "PHID-DREV-xyz",
        "wip": False,
        "tree_hash": "deadbeefcafebabe",
        "reviewers": {"granted": ["r1", "r2"], "request": ["r3"]},
    }

    assert commit_dict == expected
    assert commit_dict_loaded == expected

    # Check the `reviewers` key matches expected values.
    assert isinstance(commit_dict["reviewers"], dict)
    assert commit_dict["reviewers"]["granted"] == ["r1", "r2"]
    assert commit_dict["reviewers"]["request"] == ["r3"]


def test_commit_to_dict_defaults_are_serializable():
    # Create a commit with default values.
    commit = Commit()

    # Converting to `dict` and dumping to string should not raise.
    commit_dict = commit.to_dict()
    json.dumps(commit_dict)

    # Check a couple of defaults explicitly
    assert commit_dict["name"] == ""
    assert commit_dict["reviewers"] == {}
    assert commit_dict["rev_id"] is None
    assert commit_dict["submit"] is False
